/*
 * Copyright (C) 2006- FUJIWARA Katsunori(foozy@lares.dti.ne.jp)
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <sys/stat.h>
#include <windows.h>

/* ######################################## */

#define PATH_BUFFER_SIZE PATH_MAX

/* ######################################## */

#ifndef WRAPT_COMMAND
#error "WRAPT_COMMAND should be defined to specify wrapt cygwin command"
#define WRAPT_COMMAND ""
#endif
#ifndef SCRIPT_SUFFIX
#error "SCRIPT_SUFFIX should be defined to specify script file suffix"
#define SCRIPT_SUFFIX ""
#endif

#ifdef DEBUG
#define DEBUG_EVAL(a) a
#else
#define DEBUG_EVAL(a)
#endif

/* ######################################## */

#ifdef DEBUG

static void
showCommandline(const char* argv[])
{
    int i;
    for(i = 0 ; NULL != argv[i] ; i += 1){
        fprintf(stderr, "%s ", argv[i]);
    }
    fprintf(stderr, "\n");
}

#endif

static void
fail(const char* command,
     const char* function)
{
    fprintf(stderr,
            "%s: %s: %s\n",
            command,
            function,
            strerror(errno));
}

#ifdef OPT_ARG
/*
 * args addendum:
 * 1. interpreter(= wrapt command)
 * 2. optional argument(e.g.: "-f" for awk)
 * 3. NULL termination
 */
#define ARG_ADDENDUM 3
#else
/*
 * args addendum:
 * 1. interpreter(= wrapt command)
 * 2. NULL termination
 */
#define ARG_ADDENDUM 2
#endif

#define ARGV_SIZE(argc) (sizeof(const char*) * (argc + ARG_ADDENDUM + 1))

#define RESULT_CODE_FAIL 1

static int
requiresConversion(const char* path)
{
    /*
     * According to "man" page, isalpha() may return true
     * for non english alphabet in some locales.
     * So, this implementation does not use it.
     */
    switch(path[0]){
      case 'a': case 'b': case 'c': case 'd': case 'e': case 'f': case 'g':
      case 'h': case 'i': case 'j': case 'k': case 'l': case 'm': case 'n':
      case 'o': case 'p': case 'q': case 'r': case 's': case 't': case 'u':
      case 'v': case 'w': case 'x': case 'y': case 'z':

      case 'A': case 'B': case 'C': case 'D': case 'E': case 'F': case 'G':
      case 'H': case 'I': case 'J': case 'K': case 'L': case 'M': case 'N':
      case 'O': case 'P': case 'Q': case 'R': case 'S': case 'T': case 'U':
      case 'V': case 'W': case 'X': case 'Y': case 'Z':
        break;

      default:
        return 0;
    }

    return ((':' == path[1]) && (('/' == path[2]) || ('\\' == path[2])));
}

int
main(int argc,
     const char* argv[])
{
    DEBUG_EVAL(showCommandline(argv));

    const char* command = argv[0];
    const char* wrapt = WRAPT_COMMAND;
    const char* suffix = SCRIPT_SUFFIX;

    char script[PATH_BUFFER_SIZE];

    {
        char win32path[PATH_BUFFER_SIZE];
        char posixpath[PATH_BUFFER_SIZE];

        GetModuleFileName(NULL, win32path, sizeof(win32path)/sizeof(char));
        cygwin_conv_to_posix_path(win32path, posixpath);

        char* dot = strrchr(posixpath, '.');
        if(NULL != dot){ *dot = '\0'; }

        sprintf(script, "%s%s", posixpath, suffix);
    }

    DEBUG_EVAL(fprintf(stderr, "script=%s\n", script));
    {
        struct stat statbuf;
        if(stat(script, &statbuf)){
            fail(command, "stat");
            return RESULT_CODE_FAIL;
        }
        if(0 == S_ISREG(statbuf.st_mode)){
            fprintf(stderr,
                    "%s: %s is not regular file\n",
                    command,
                    script);
            return 1;
        }
    }

    const char** invokee_argv = (const char**)(malloc(ARGV_SIZE(argc)));
    if(NULL == invokee_argv){
        fail(command, "malloc");
        return RESULT_CODE_FAIL;
    }

    int index = -1;
    invokee_argv[(index += 1)] = wrapt;
#ifdef OPT_ARG
    invokee_argv[(index += 1)] = OPT_ARG;
#endif
    invokee_argv[(index += 1)] = script;

    int i;
    for(i = 1 ; i < argc ; i += 1){
        if(requiresConversion(argv[i])){
            char* path = (char*)(malloc(PATH_BUFFER_SIZE));
            if(NULL == path){
                fail(command, "malloc");
                return RESULT_CODE_FAIL;
            }

            cygwin32_conv_to_posix_path(argv[i], path);
            invokee_argv[(index += 1)] = path;
        }
        else{
            invokee_argv[(index += 1)] = argv[i];
        }
    }
    invokee_argv[(index += 1)] = NULL;
    DEBUG_EVAL(showCommandline(invokee_argv));

    execvp(wrapt, (char* const*)invokee_argv);
    /*
     * above does not return control to here, if "execv" successes.
     */

    fail(command, "execv");
    return RESULT_CODE_FAIL;
}
